# -*- coding: utf-8 -*-

from django.db import models, transaction
from django.contrib.auth.models import User
from django.utils.translation import ugettext_lazy as _
from django.conf import settings
import datetime

TASK_STATE_SCHEDULED = 0
TASK_STATE_FAILED = 1
TASK_STATE_HOLD = 2
TASK_STATE_ABORTED = 3
TASK_STATE_COMPLETED = 4
TASK_STATE_CANCELED = 5
TASK_STATE_RUNNING = 6

TASK_STATES = ((TASK_STATE_SCHEDULED, _(u'programmée')),
               (TASK_STATE_RUNNING, _(u'en cours')),
               (TASK_STATE_FAILED, _(u'échouée temporairement')),
               (TASK_STATE_HOLD, _(u'en attente')),
               (TASK_STATE_ABORTED, _(u'échouée définitivement')),
               (TASK_STATE_COMPLETED, _(u'terminée')),
               (TASK_STATE_CANCELED, _(u'annulée')),
           )

class TaskManager(models.Manager):
    def pending(self):
        """
        Tasks that are pending
        """
        return self.filter(state__in = (TASK_STATE_SCHEDULED, TASK_STATE_FAILED))

    def pop(self):
        """
        Pop the first action to execute, or return None
        """
        tasks = self.pending().order_by("-priority","created_on")
        for task in tasks:
            if task.available():
                return task
        return None
        

class Task(models.Model):
    class Meta:
        verbose_name = _(u'tâche')

    objects = TaskManager()

    # Task metadata
    created_on = models.DateTimeField(_(u"date de création"),
                                        auto_now_add = True, db_index = True)
    processed_on = models.DateTimeField(_(u"date de réalisation"),
                                        blank = True, null = True)
    progress = models.IntegerField(_(u"état d'avancement"), null = True,
                                   blank = True)
    nb_tries = models.IntegerField(_(u"nombre de tentatives maximales"),
                                   default = 1)
    cur_tries = models.IntegerField(_(u"nombre de tentatives actuelles"),
                                    default = 0)
    state = models.IntegerField(_(u"état de la tâche"),
                                default = 0, choices = TASK_STATES,
                                db_index = True)    
    blockers = models.ManyToManyField("Task", verbose_name = _(u"bloquée par"),
                                      blank = True)
    run_on = models.DateTimeField(_(u"date cible"),
                                  blank = True, null = True)

    priority = models.IntegerField(_(u"priorité"), default = 0)

    # Task data
    name = models.CharField(_(u"nom"), max_length = 250)
    parameters = models.TextField(_(u"paramètres"))
    result = models.TextField(_(u"résultat"), blank = True, null = True)
    owner = models.ForeignKey(User, verbose_name=u"Propriétaire", null=True, blank=True)

    def __unicode__(self):
        return u"%d - %s - %s - %s" % (self.id, self.name, 
                                       self.created_on, self.get_state_display())

    def available(self):
        """
        Check if the task is currently available
        """
        return not (self.is_waiting_cooldown() or self.is_blocked() or self.is_delayed())

    def is_blocked(self):
        """
        Check if the task is currently blocked
        """
        return self.blockers.exclude(state = TASK_STATE_COMPLETED).count()

    def is_waiting_cooldown(self):
        """
        Check if the task is waiting the cooldown (ie, it failed, must be
        retried, but it's too soon yet)
        """
        cooldown = getattr(settings, 'LEGUME_COOLDOWN', 120)
        if self.state == TASK_STATE_FAILED and self.processed_on:
            when = self.processed_on + datetime.timedelta(seconds=cooldown)
            return when > datetime.datetime.now()
        return False

    def is_delayed(self):
        """
        Check if the task is delayed
        """
        return self.run_on and self.run_on > datetime.datetime.now()

    @transaction.commit_manually
    def update_progress(self, progress):
        self.progress = progress
        self.save()
        transaction.commit()
