# -*- coding: utf-8 -*-

from django.contrib.auth.models import User
from legume.models import *
from django.conf import settings
import cPickle as pickle
import datetime
from django.db import transaction

from loggers import logger
from functools import wraps

TASK_REGISTRY = {}
PERIODIC_TASK_REGISTRY = {}

class LegumeTask(object):
    """
    This is the real task object, wrapping the decorated function
    """
    def __init__(self, callback, options = {}):
        """
        Constructor, fills the registry
        """
        self.callback = callback
        self.options = dict(options)

        if not "nb_tries" in self.options:
            self.options["nb_tries"] = getattr(settings, "LEGUME_NB_TRIES", 1)

        self.options["priority"] = int(self.options.get("priority", 0))

        self.provide_task = self.options.get("provide_task", False)
        self.name = self.options.get("name", None)
        self.task = None
        
        if not self.name:
            self.name = "%s.%s" % (callback.__module__, callback.__name__)

        if self.name in TASK_REGISTRY:
            raise ValueError, "cannot have two tasks of the same name"
        TASK_REGISTRY[self.name] = self

    def delay(self, request = None, *args, **kwargs):
        """
        Program the task for later execution
        """
        return self.delay_after(request, [], *args, **kwargs)

    def delay_after(self, request = None, blockers = [], *args, **kwargs):
        """
        Program the task for later execution, after given blockers
        """
        parameters = pickle.dumps((args, kwargs)).encode('base64')
        return self.schedule(blockers, parameters, request = request)

    def schedule(self, blockers, parameters, request = None, **extra):
        """
        Program the task for later execution, after given blockers
        """
        task = Task(name = self.name,
                    parameters = parameters,
                    nb_tries = self.options['nb_tries'],
                    priority = self.options["priority"],
                    **extra)
        user = getattr(request, 'user', None)
        if isinstance(user, User) and not user.is_anonymous():
            task.owner = user
        if blockers:
            task.state = TASK_STATE_HOLD
        task.save()
        if blockers:
            task.blockers = blockers
            task.state = TASK_STATE_SCHEDULED
            task.save()
        return task

    def run(self, task):
        """
        Run the given scheduled task
        """
        self.task = task
        args, kwargs = pickle.loads(task.parameters.decode('base64'))
        if self.provide_task:
            args = (task,) + tuple(args)
        result = self(*args, **kwargs)
        task.result = pickle.dumps(result).encode('base64')

    def __call__(self, *args, **kwargs):
        """
        Just call the callback
        """
        return self.callback(*args, **kwargs)

    def get_logger(self):
        """
        Wrapper to get a logger
        """
        return logger

    def get_task(self):
        """
        Wrapper to get the task
        """
        return self.task

class LegumePeriodicTask(LegumeTask):
    """
    A periodic task
    """
    def __init__(self, run_every, run_at, callback, options = {}):
        """
        Constructor
        """
        if not "nb_tries" in options:
            options["nb_tries"] = 1 # By default, don't retry periodic tasks
        options["priority"] = int(options.get("priority", 0))
        super(LegumePeriodicTask, self).__init__(callback, options)
        self.run_every = run_every
        self.run_at = run_at
        PERIODIC_TASK_REGISTRY[self.name] = self
        
    def schedule(self, blockers, parameters, request = None, **extra):
        """
        Schedule the task in the future, if not already existing
        """
        already = Task.objects.pending().filter(name = self.name)
        if already.count():
            logger.debug("Task %s already scheduled" % self.name)
            return

        now = datetime.datetime.now()
        if self.run_every:
            run_on = now + self.run_every
        elif self.run_at:
            if now.time() > self.run_at:
                # Too late for today ? well, ok, do it tomorrow
                now += datetime.timedelta(days=1)
            run_on = datetime.datetime.combine(now, self.run_at)

        logger.info("Scheduling %s on %s" % (self.name, run_on))
        return super(LegumePeriodicTask, self).schedule(blockers, parameters,
                                                        run_on = run_on, **extra)

def task(*args, **kwargs):
    """
    Use this as a decorator :
    @task

    Or with named parameters :
    @task(name = 'my task', nb_tries = 10)
    """
    if len(args) and len(kwargs):
        raise ValueError, "task cannot be called with both named and unnamed parameters"
    if len(kwargs):
        options = dict(kwargs)
        def make_task(callback):
            return LegumeTask(callback, options)
        return make_task

    if len(args) != 1:
        raise ValueError, "task takes exactly one unnamed argument"

    callback = args[0]
    return LegumeTask(callback)

def periodic_task(run_every = None, run_at = None, **kwargs):
    """
    Use this as a decorator with arguments:
    @periodic_task(run_every = timedelta(hours=24), name = 'my dialy task')
    or (for daily tasks runned at fixed time) :
    @periodic_task(run_at = time(23, 30), name = 'my nightly task')
    """
    if (not run_every and not run_at) or (run_every and run_at):
        raise ValueError, "exactly one of run_every or run_at is required for periodic_task"

    options = dict(kwargs)
    def make_task(callback):
        return LegumePeriodicTask(run_every, run_at, callback, options)
    return make_task

@transaction.commit_manually
def run_task(task):
    """
    Run a programmed task object
    """
    logger.info("Running %s" % task)
    task.cur_tries += 1
    task.processed_on = datetime.datetime.now()
    task.state = TASK_STATE_RUNNING
    task.save()
    transaction.commit()

    try:
        ltask = TASK_REGISTRY[task.name]
        ltask.run(task)
        transaction.commit()
    except Exception, e:
        transaction.rollback()
        if task.cur_tries < task.nb_tries:            
            logger.warning("Got error %s in %s, will try again" % (e, task))
            task.state = TASK_STATE_FAILED
        else:
            logger.exception("Got final error in %s" % task)
            task.state = TASK_STATE_ABORTED
    else:
        logger.info("Done running %s" % task)
        task.state = TASK_STATE_COMPLETED
    task.save()
    transaction.commit()

def schedule_periodic_tasks():
    """
    Schedule the next batch of periodic tasks, if needed
    """
    for task in PERIODIC_TASK_REGISTRY.values():
        task.delay()
